package de.mbutscher.wikiandpad;

import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.mbutscher.wikiandpad.db.DbConsts;
import de.mbutscher.wikiandpad.db.DbReadAccessException;
import de.mbutscher.wikiandpad.db.DbWriteAccessException;
import de.mbutscher.wikiandpad.db.IWikiData;
import de.mbutscher.wikiandpad.db.MatchTermEntry;
import de.mbutscher.wikiandpad.db.TimestampType;
import de.mbutscher.wikiandpad.db.WikiDataException;
import de.mbutscher.wikiandpad.defaultwikilang.WikiLangParser;
import de.mbutscher.wikiandpad.parsing.SyntaxNode;
import de.mbutscher.wikiandpad.utils.MiscEvent;
import de.mbutscher.wikiandpad.utils.MiscEventSource;
import de.mbutscher.wikiandpad.utils.NotCurrentThreadException;
import de.mbutscher.wikiandpad.utils.ThreadHolder;

public class WikiPage implements MiscEventSource {
    
    protected final ThreadHolder livePageAstUpdate = new ThreadHolder();
    protected WikiDocument wikiDocument;
    protected String wikiPageName;
    protected String editorText;

    protected SyntaxNode livePageAst;
    protected String liveTextForPageAst;
    
    protected MiscEvent miscEvent = new MiscEvent(this);


    public WikiPage(WikiDocument wikiDocument, String wikiPageName) {
        this.wikiDocument = wikiDocument;
        this.wikiPageName = wikiPageName;
    }
    
    public MiscEvent getMiscEvent() {
        return miscEvent;
    }

    /**
     * Return the name of the unified name of the page, which is
        "wikipage/" + the wiki word for wiki pages or the functional tag
        for functional pages.
     */
    public String getUnifiedPageName() {
        return "wikipage/" + wikiPageName;
    }
    
    public String getWikiPageName() {
        return wikiPageName;
    }
    
    
    /**
     * Return true if page is effectively read-only, this means
        "for any reason", regardless if error or intention.
     * @return
     */
    public boolean isReadOnlyEffect() {
        return wikiDocument == null || wikiDocument.isReadOnlyEffect();
    }
    
    
    public void invalidate() {
        wikiDocument = null;
    }
    
    public boolean isValid() {
        return wikiDocument != null;
    }

    public IWikiData getWikiData() {
        return wikiDocument.getWikiData();
    }

    public String getWikiPageContent()
            throws DbReadAccessException {
        String result = wikiDocument.getUnifiedPageContent(
                getUnifiedPageName());
        
        if (result == null)
            return "";
        
        return result;
    }
    
    public WikiDocument getWikiDocument() {
        return wikiDocument;
    }

    public synchronized void writeToDatabase(String text, boolean fireEvent)
            throws DbWriteAccessException {
        if (isReadOnlyEffect())
            return;
        
        wikiDocument.setWikiPageContent(wikiPageName, text);
        refreshSyncUpdateMatchTerms();
        initiateUpdate(fireEvent);
    }

    public String getLiveText() throws DbReadAccessException {
        if (editorText != null)
            return editorText;
        
        return getWikiPageContent();
    }
    
    
    public void setLiveText(String liveText) throws DbWriteAccessException {
        if (editorText != null)
            setEditorText(liveText);
        else
            writeToDatabase(liveText, true);
            
    }

    
    public String getEditorText() {
        return editorText;
    }

    public void setEditorText(String editorText) {
        this.editorText = editorText;
        initiateLivePageAstUpdate();
    }
    
    protected synchronized void initiateLivePageAstUpdate() {
        final String currentLiveText = editorText;
        if (currentLiveText == null) {
            livePageAstUpdate.setThread(null);
            return;
        }
        
        livePageAstUpdate.execute(new Runnable() {
            public void run() {
                try {
                    SyntaxNode pageAst = WikiLangParser.parseText(
                            currentLiveText, WikiPage.this, livePageAstUpdate);
                    setLivePageAst(pageAst, currentLiveText);
                } catch (NotCurrentThreadException e) {
                    return;
                }
            }
        });
    }
    
    protected synchronized void setLivePageAst(SyntaxNode livePageAst,
            String liveTextForPageAst) throws NotCurrentThreadException {
        livePageAstUpdate.testValidThread();
        this.livePageAst = livePageAst;
        this.liveTextForPageAst = liveTextForPageAst;
        notifyAll();
        getMiscEvent().cloneIt().put("updated livePageAst")
                .put("livePageAst", livePageAst)
                .put("liveTextForPageAst", liveTextForPageAst)
                .send();
    }
    
    public synchronized SyntaxNode getLivePageAstIfAvailable() {
        if (editorText == liveTextForPageAst)
            return livePageAst;
        else
            return null;
    }
    
    public synchronized SyntaxNode getLivePageAst() {
        while (editorText != liveTextForPageAst) {
            if (livePageAstUpdate.getThread() == null)
                initiateLivePageAstUpdate();

            try {
                wait();
            } catch (InterruptedException e) {
            }
        }
        
        return livePageAst;
    }
    


    protected void initiateUpdate(boolean fireEvent) {
        // TODO Auto-generated method stub
    }
    
    /**
     * Refresh those match terms which must be refreshed synchronously
     */
    public void refreshSyncUpdateMatchTerms() {
        if (wikiDocument.isReadOnlyEffect())
            return;
        
        int WORD_TYPE = DbConsts.WIKIWORDMATCHTERMS_TYPE_ASLINK | 
                DbConsts.WIKIWORDMATCHTERMS_TYPE_FROM_WORD | 
                DbConsts.WIKIWORDMATCHTERMS_TYPE_SYNCUPDATE;
        
        List<MatchTermEntry> matchTerms = Arrays.asList(new MatchTermEntry[]
                {new MatchTermEntry(wikiPageName, WORD_TYPE, wikiPageName, -1,
                0, 0)});
        
        wikiDocument.updateWikiWordMatchTerms(wikiPageName, matchTerms, true);
    }
    
    /**
     * This is called by WikiDocument during
        {@link WikiDocument#renameWikiWord(String, String, boolean)}
        and shouldn't be called elsewhere.
     * @param toWord
     * @throws DbWriteAccessException 
     */
    public void renameVersionData(String toWord)
            throws DbWriteAccessException {

        if (isReadOnlyEffect())
            return;

        IWikiData wikiData = getWikiData();

        wikiData.beginSingleTransaction();
        try {        
            wikiData.renameDataBlock("versioning/overview/" + wikiPageName,
                    "versioning/overview/" + toWord);
            
            Pattern pattern = Pattern.compile(
                    "(versioning/packet/versionNo/[0-9]+/)" + Pattern.quote(
                    wikiPageName) + "$");

            List<String> candidates = wikiData
                    .getDataBlockUnifNamesStartingEndingWith(
                            "versioning/packet/versionNo/", "/" + wikiPageName);
            
            for(String c: candidates) {
                Matcher matcher = pattern.matcher(c);
                if (!matcher.matches())
                    continue;
                wikiData.renameDataBlock(c, matcher.group(1) + toWord);
            }

            wikiData.setSingleTransactionSuccessful();
            wikiDocument.newDatabaseChangeCount();
        }
        finally {
            wikiData.endSingleTransaction();
        }

    }
    
    /**
     * Deletes the page from database
     * @param fireEvent
     * @throws WikiDataException 
     */
    public synchronized void deletePage(boolean fireEvent)
            throws WikiDataException {
        if (isReadOnlyEffect())
            return;

        IWikiData wikiData = getWikiData();

        wikiData.beginSingleTransaction();
        try {        

            if (wikiData.isDefinedWikiPageName(wikiPageName))
                    wikiData.deleteWikiPage(wikiPageName);

            // Delete versioning data
            wikiData.deleteDataBlock("versioning/overview/" + wikiPageName);

            Pattern pattern = Pattern.compile(
                    "(versioning/packet/versionNo/[0-9]+/)" + Pattern.quote(
                            wikiPageName) + "$");

            List<String> candidates = wikiData
                    .getDataBlockUnifNamesStartingEndingWith(
                            "versioning/packet/versionNo/", "/" + wikiPageName);

            for(String c: candidates) {
                Matcher matcher = pattern.matcher(c);
                if (!matcher.matches())
                    continue;
                wikiData.deleteDataBlock(c);
            }

            wikiData.setSingleTransactionSuccessful();
            wikiDocument.newDatabaseChangeCount();
            invalidate();        
        }
        finally {
            wikiData.endSingleTransaction();
        }
    }
    
    /**
     * Called to inform the page that it was visited and should set
     * the "visited" entry in the database to current time.
     */
    public void informVisited() {
        if (isReadOnlyEffect())
            return;

        if (!isDefined())
            return;

        IWikiData wikiData = getWikiDocument().getWikiData();
        wikiData.setTimestamp(getWikiPageName(), TimestampType.VISITED,
                System.currentTimeMillis() / 1000.0);

        getWikiDocument().newDatabaseChangeCount();
    }

    
    public boolean isDefined() {
        return getWikiDocument().isDefinedWikiPageName(getWikiPageName());
    }

}


/*

    def deletePage(self, fireEvent=True):
        """
        Deletes the page from database
        """
        with self.textOperationLock:
            if self.isReadOnlyEffect():
                return
    
            if self.isDefined():
                self.getWikiData().deleteWord(self.getWikiWord())

            vo = self.getExistingVersionOverview()
            if vo is not None:
                vo.delete()
                self.versionOverview = UNDEFINED
        
            self.queueRemoveFromSearchIndex()   # TODO: Check for (dead-)locks

            if fireEvent:
                wx.CallAfter(self.fireMiscEventKeys,
                        ("deleted page", "deleted wiki page"))




*/







