package de.mbutscher.wikiandpad.db.compactSqlite;

import java.io.File;
import java.io.UnsupportedEncodingException;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import de.mbutscher.wikiandpad.BadProgrammingException;
import de.mbutscher.wikiandpad.WikiDocument;
import de.mbutscher.wikiandpad.db.ConnectWrapSqlite;
import de.mbutscher.wikiandpad.db.DbReadAccessException;
import de.mbutscher.wikiandpad.db.DbWriteAccessException;
import de.mbutscher.wikiandpad.db.IWikiData;
import de.mbutscher.wikiandpad.db.MatchTermEntry;
import de.mbutscher.wikiandpad.db.TimestampType;
import de.mbutscher.wikiandpad.db.WikiDataException;
import de.mbutscher.wikiandpad.utils.Utilities;
import android.annotation.SuppressLint;
import android.database.sqlite.SQLiteDatabase;

public class WikiData implements IWikiData {
    protected String wikiDbPath;
    protected String wikiConfigPath;
    protected String wikiDataDirPath;
    protected SQLiteDatabase database;
    protected ConnectWrapSqlite connWrap;
    protected WikiDocument wikiDocument;

    // Supported versions:
    // Version 2 = WikidPad 2.0
    // Version 3 = WikidPad 2.1-
    protected int writeCompatVer;
    
    public WikiData(String wikiConfigPath, WikiDocument wikiDocument)
            throws WikiDataException {
        this.wikiConfigPath = wikiConfigPath;
        this.wikiDocument = wikiDocument;
        
        File fTemp = (new File(wikiConfigPath)).getParentFile();
        fTemp = new File(fTemp, "data");
        wikiDataDirPath = fTemp.getAbsolutePath();

        fTemp = new File(fTemp, "wiki.sli");
        wikiDbPath = fTemp.getAbsolutePath();
        database = SQLiteDatabase.openDatabase(wikiDbPath, null,
                SQLiteDatabase.OPEN_READWRITE |
                SQLiteDatabase.NO_LOCALIZED_COLLATORS);

        connWrap = new ConnectWrapSqlite(database);
        
        writeCompatVer = connWrap.getSettingsInt("writecompatver", -1);
        if (writeCompatVer < 8 || writeCompatVer > 9) {
            throw new WikiDataException(String.format(
                    "Database version %s not supported", writeCompatVer));
        }
    }
    
    public String getWikiConfigPath() {
        return wikiConfigPath;
    }

    public synchronized void close() {
        if (connWrap == null)
            return;
        
        connWrap.close();
        connWrap = null;
        database = null;
        wikiDbPath = null;
    }
    

    public void beginSingleTransaction() {
        connWrap.beginSingleTransaction();
    }

    public void endSingleTransaction() {
        connWrap.endSingleTransaction();
    }

    public void setSingleTransactionSuccessful() {
        connWrap.setSingleTransactionSuccessful();
    }

    public synchronized String getWikiPageContent(String wikiPageName)
            throws DbReadAccessException {
        byte[] content = connWrap.execSqlQuerySingleItem("select content from "+
                "wikiwordcontent where word = ?", wikiPageName);
        if (content == null) {
            return null;
        }
        
        try {
            return new String(content, "UTF-8");
        } catch (UnsupportedEncodingException e) {
            return null;
        }
    }

    
    /**
     * Sets the content, does not modify the cache information
        except cachedWikiPageLinkTermDict
     * @param wikiPageName
     * @param content
     * @throws DbWriteAccessException 
     */
    public synchronized void setWikiPageContent(String wikiPageName,
            String strContent, double moddate, double creadate)
            throws DbWriteAccessException {
    
        double ti = System.currentTimeMillis() / 1000.0;
        if (moddate < 0)
            moddate = ti;
        
        String data = connWrap.execSqlQuerySingleItem(
                "select word from wikiwordcontent where word = ?",
                wikiPageName);
        
        byte[] content;
        try {
            content = strContent.getBytes("UTF-8");
        } catch (UnsupportedEncodingException e) {
            throw new BadProgrammingException("Encoding UTF-8 unknown");
        }
        
        connWrap.beginSingleTransaction();
        try {
            if (data == null) {
                if (creadate < 0)
                    creadate = ti;
                
                connWrap.execSql("insert into wikiwordcontent(word, created, " +
                        "modified, content) values (?, ?, ?, ?)",
                        wikiPageName, creadate, moddate, content);
            }
            else {
                connWrap.execSql("update wikiwordcontent set modified = ?, " +
                        "content = ? where word = ?", moddate, content,
                        wikiPageName);
            }
    
            connWrap.execSql("update wikiwordcontent set " + 
                    "metadataprocessed = 0 where word = ?", wikiPageName);
    
    
            connWrap.setSingleTransactionSuccessful();
            wikiDocument.newDatabaseChangeCount();
        }
        finally {
            connWrap.endSingleTransaction();
        }
    }

    public synchronized boolean isDefinedWikiPageName(String wikiPageName) {
        return connWrap.execSqlQuerySingleItem(
                "select word from wikiwordcontent where word = ?",
                wikiPageName) != null;
    }
    
    
    
    /**
     * The content which was stored under wikiPageName is stored
        after the call under toWord.
     * @param wikiPageName
     * @param toWord
     * @throws DbWriteAccessException 
     */
    protected void _renameContent(String wikiPageName,
            String toWord) throws DbWriteAccessException {
    
        connWrap.execSql("update wikiwordcontent set word = ?, " +
                "metadataprocessed = 0 where word = ?",
                toWord, wikiPageName);
    }

    public synchronized void renameWord(String wikiPageName, String toWord)
            throws DbWriteAccessException {
        connWrap.beginSingleTransaction();
        try {        
    
            connWrap.execSql("update wikirelations set word = ? where word = ?",
                    toWord, wikiPageName);
            if (writeCompatVer == 8) {
                connWrap.execSql("update wikiwordprops set word = ? where " +
                        "word = ?", toWord, wikiPageName); 
            }
            else {
                connWrap.execSql("update wikiwordattrs set word = ? where " +
                        "word = ?", toWord, wikiPageName);
            }
            connWrap.execSql("update todos set word = ? where word = ?",
                    toWord, wikiPageName);
            connWrap.execSql("update wikiwordmatchterms set word = ? where " +
                    "word = ?", toWord, wikiPageName);
            _renameContent(wikiPageName, toWord);
    
            connWrap.setSingleTransactionSuccessful();
            wikiDocument.newDatabaseChangeCount();
        }
        finally {
            connWrap.endSingleTransaction();
        }
    }

    /**
     * delete everything about the wikiword passed in. an exception is raised
        if you try and delete the wiki root node.
    
     * @param wikiPageName
     * @param delContent Should actual content be deleted as well? (Parameter is
                not part of official API)
     * @throws WikiDataException 
     */
    public synchronized void deleteWikiPage(String wikiPageName,
            boolean delContent)
            throws WikiDataException {
        if (wikiPageName.equals(wikiDocument.getWikiName()))
            throw new WikiDataException("You cannot delete the root wiki node");
        
        connWrap.beginSingleTransaction();
        try {
            /*
            don't delete the relations to the word since other
            pages still have valid outward links to this page.
            just delete the content
            */
            
            deleteChildRelationships(wikiPageName);
            deleteAttributes(wikiPageName);
            deleteTodos(wikiPageName);
            if (delContent)
                _deleteContent(wikiPageName);
            deleteWikiWordMatchTerms(wikiPageName, false);
            deleteWikiWordMatchTerms(wikiPageName, true);
            
        
            connWrap.setSingleTransactionSuccessful();
            wikiDocument.newDatabaseChangeCount();
        }
        finally {
            connWrap.endSingleTransaction();
        }
    }

    public synchronized void deleteWikiPage(String wikiPageName)
            throws WikiDataException {
        deleteWikiPage(wikiPageName, true);
    }

    /**
     * get the list of match terms with thisStr in them.
     * @param thisStr
     * @param orderBy
     * @param descend
     * @return
     */
    @SuppressLint("DefaultLocale")
    public synchronized ArrayList<MatchTermEntry> getWikiWordMatchTermsWith(
            String thisStr, String orderBy, boolean descend) {
        String s1 = thisStr.toLowerCase();
        thisStr = ConnectWrapSqlite.escapeForGlob(s1);
    
        Object[][] result1;
        Object[][] result2;
        
        // charlength is supported since DB version 3
        if ("visited".equals(orderBy)) {
            result1 = connWrap.execSqlQuery(
                    "select matchterm, type, wikiwordmatchterms.word, " +
                    "firstcharpos, -1, visited " +
                    // "firstcharpos, charlength, visited " +
                    "from wikiwordmatchterms inner join wikiwordcontent " +
                    "on wikiwordmatchterms.word = wikiwordcontent.word where " +
                    "matchtermnormcase glob (? || '*')", thisStr);
            
            result2 = connWrap.execSqlQuery(
                    "select matchterm, type, wikiwordmatchterms.word, " +
                    // "firstcharpos, charlength, visited " +
                    "firstcharpos, -1, visited " +
                    "from wikiwordmatchterms inner join wikiwordcontent " +
                    "on wikiwordmatchterms.word = wikiwordcontent.word where " +
                    "not matchtermnormcase glob (? || '*') " +
                    "and matchtermnormcase glob ('*' || ? || '*')",
                    thisStr, thisStr);
        }
        else {
            result1 = connWrap.execSqlQuery(
                    // "select matchterm, type, word, firstcharpos, charlength " +
                    "select matchterm, type, word, firstcharpos, -1 " +
                    "from wikiwordmatchterms where " +
                    "matchtermnormcase glob (? || '*')", thisStr);
    
            result2 = connWrap.execSqlQuery(
                    "select matchterm, type, word, firstcharpos, -1 " +
                    // "select matchterm, type, word, firstcharpos, charlength " +
                    "from wikiwordmatchterms where " +
                    "not matchtermnormcase glob (? || '*') " +
                    "and matchtermnormcase glob ('*' || ? || '*')",
                    thisStr, thisStr);
        }
        
        final Collator coll = wikiDocument.getCollator();
        final Comparator<Object[]> compCollFirst = new Comparator<Object[]> () {
            public int compare(Object[] arg0, Object[] arg1) {
                return coll.compare(arg0[0], arg1[0]);
            }
        };
        
        Arrays.sort(result1, compCollFirst);
        Arrays.sort(result2, compCollFirst);
        
        if ("visited".equals(orderBy)) {
            final Comparator<Object[]> compVisited = new Comparator<Object[]> () {
                public int compare(Object[] arg0, Object[] arg1) {
                    try {
                        double c = ((Double)arg1[5]) - ((Double)arg0[5]);
                        
                        if (c == 0)
                            return 0;
                        if (c > 0)
                            return 1;
                        
                        return -1;
                    }
                    catch(Exception e) {
                        // TODO: Error handling
                        return compCollFirst.compare(arg0, arg1);
                    }
                }
            };
            
            Arrays.sort(result1, compVisited);
            Arrays.sort(result2, compVisited);
        }
        
    
        ArrayList<MatchTermEntry> result = new ArrayList<MatchTermEntry>();
                
        for (Object[] e: result1) {
            result.add(new MatchTermEntry(e));
        }
    
        for (Object[] e: result2) {
            result.add(new MatchTermEntry(e));
        }
    
        return result;
    }

    @SuppressLint("DefaultLocale")
    protected synchronized void _addWikiWordMatchTerm(MatchTermEntry mte) {
        // TODO Check for name collisions
        
        if (writeCompatVer < 3) {
            connWrap.execSql("insert into wikiwordmatchterms(matchterm, " +
                    "type, word, firstcharpos, matchtermnormcase) " +
                    "values (?, ?, ?, ?, ?)",
                    mte.matchterm, mte.type, mte.wikiPageName, mte.firstcharpos,
                    mte.matchterm.toLowerCase());
        }
        else if (writeCompatVer > 2) {
            connWrap.execSql("insert into wikiwordmatchterms(matchterm, " +
                    "type, word, firstcharpos, charlength, matchtermnormcase) " +
                    "values (?, ?, ?, ?, ?, ?)",
                    mte.matchterm, mte.type, mte.wikiPageName, mte.firstcharpos,
                    mte.charlength, mte.matchterm.toLowerCase());
        }
    }

    public synchronized void updateWikiWordMatchTerms(String wikiPageName,
            List<MatchTermEntry> wwmTerms, boolean syncUpdate) {

        connWrap.beginSingleTransaction();
        try {        
            deleteWikiWordMatchTerms(wikiPageName, syncUpdate);
            // TODO:  self.getExistingWikiWordInfo(word)  (check if wikiPageName exists)
            for(MatchTermEntry mte: wwmTerms) {
                _addWikiWordMatchTerm(mte);
            }
            connWrap.setSingleTransactionSuccessful();
            wikiDocument.newDatabaseChangeCount();
        }
        finally {
            connWrap.endSingleTransaction();
        }
    }
    
    protected void deleteWikiWordMatchTerms(String wikiPageName,
            boolean syncUpdate) {
        String addSql;
        if (syncUpdate)
            addSql = " and (type & 16) != 0";
        else
            addSql = " and (type & 16) == 0";
        // DbConsts.WIKIWORDMATCHTERMS_TYPE_SYNCUPDATE == 16
        connWrap.execSql("delete from wikiwordmatchterms where " +
                "word = ?" + addSql, wikiPageName);
    }

    /**
     * Return all unified names starting with startingWith (case sensitive)
     * @param startingWith
     * @return
     */
    public List<String> getDataBlockUnifNamesStartingWith(
            String startingWith) {
        startingWith = ConnectWrapSqlite.escapeForGlob(startingWith);
        
        Object[] data = connWrap.execSqlQuerySingleColumn(
                "select distinct(unifiedname) from datablocks where " +
                "unifiedname glob (? || '*')", startingWith);
        
        return Utilities.castList(data);
    }

    /**
     * Return all unified names starting with startingWith (case sensitive)
     * and ending with endingWith
     * @param startingWith
     * @param endingWith
     * @return
     */
    public List<String> getDataBlockUnifNamesStartingEndingWith(
            String startingWith, String endingWith) {
        startingWith = ConnectWrapSqlite.escapeForGlob(startingWith);
        endingWith = ConnectWrapSqlite.escapeForGlob(endingWith);

        Object[] data = connWrap.execSqlQuerySingleColumn(
                "select distinct(unifiedname) from datablocks where " +
                "unifiedname glob (? || '*' || ?)",
                startingWith, endingWith);
        
        return Utilities.castList(data);
    }


    /**
     * Tries to rename datablock with name unifName to toUnifName.
     * If datablock unifName doesn't exist, nothing happens.
     * @param unifName
     * @param toUnifName
     * @throws DbWriteAccessException
     */
    public synchronized void renameDataBlock(String unifName, String toUnifName)
            throws DbWriteAccessException {
        connWrap.beginSingleTransaction();
        try {        
    
            Object t = connWrap.execSqlQuerySingleItem(
                    "select unifiedname from datablocks where unifiedname = ?",
                    unifName);

            if (t != null) {
                // internal datablock
                connWrap.execSql("update datablocks set unifiedname = ? where " +
                        "unifiedname = ?",  toUnifName, unifName);
                
                connWrap.setSingleTransactionSuccessful();
                wikiDocument.newDatabaseChangeCount();
                return;
            }
        }
        finally {
            connWrap.endSingleTransaction();
        }
    }
    
    /** 
     * Delete data block with the associated unified name. If the unified name
        is not in database, nothing happens.
     * @param unifName
     */
    public synchronized void deleteDataBlock(String unifName) {
        connWrap.beginSingleTransaction();
        try {
            connWrap.execSql(
                    "delete from datablocks where unifiedname = ?", unifName);
            
            connWrap.setSingleTransactionSuccessful();
            wikiDocument.newDatabaseChangeCount();
        }
        finally {
            connWrap.endSingleTransaction();
        }
    }
    
    
    protected void _deleteContent(String wikiPageName)
            throws DbWriteAccessException {
        connWrap.execSql("delete from wikiwordcontent where word = ?",
                wikiPageName);
    }
    
    public synchronized void deleteTodos(String wikiPageName) {
        connWrap.execSql("delete from todos where word = ?", wikiPageName);
    }

    public synchronized void deleteAttributes(String wikiPageName) {
        if (writeCompatVer == 8) {
            connWrap.execSql("delete from wikiwordprops where word = ?",
                    wikiPageName);            
        }
        else {
            connWrap.execSql("delete from wikiwordattrs where word = ?",
                    wikiPageName);
        }
    }
    
    public synchronized void deleteChildRelationships(String wikiPageName) {
        connWrap.execSql("delete from wikirelations where word = ?",
                wikiPageName);
    }
    
    public synchronized String getWikiPageNameForLinkTerm(String linkTerm) {
        return connWrap.execSqlQuerySingleItem(
                "select word " +
                "from wikiwordmatchterms where " +
                "(type & 2) != 0 and " +
                "matchterm = ?", linkTerm);

        // Consts.WIKIWORDMATCHTERMS_TYPE_ASLINK == 2
    }
    
    /**
        Returns the name of the "first" wiki word. See getNextWikiPageName()
        for details. Returns either an existing wiki word or None if no
        wiki words in database.
        Function must work for read-only wiki.
     */
    public synchronized String getFirstWikiPageName() {
        return connWrap.execSqlQuerySingleItem(
                "select word from wikiwordcontent " +
                "order by word limit 1");
    }


    /**
        Returns the "next" wiki word after currWord or None if no
        next word exists. If you begin with the first word returned
        by getFirstWikiPageName() and then use getNextWikiPageName() to
        go to the next word until no more words are available
        and if the list of existing wiki words is not modified during
        iteration, it is guaranteed that you have visited all real
        wiki words (no aliases) then.
        currWord  doesn't have to be an existing word itself.
        Function must work for read-only wiki.

     * @return
     */
    public synchronized String getNextWikiPageName(String currWord) {
        return connWrap.execSqlQuerySingleItem(
                "select word from wikiwordcontent where " +
                "word > ? order by word limit 1", currWord);
    }

    protected String mapTimestampTypeToColumnName(TimestampType tt) {
        switch (tt) {
        case CREATED: return "created";
        case MODIFIED: return "modified";
        case VISITED: return "visited";
        }
        return null;
    }
    
    public double getTimestamp(String wikiPageName, TimestampType tt) {
        String colName = mapTimestampTypeToColumnName(tt);
        
        Double t = connWrap.execSqlQuerySingleItem(
                "select " + colName + " from wikiwordcontent " +
                "where word = ?", wikiPageName);
        
        if (t == null)
            return 0;
        
        return t;
    }

    public void setTimestamp(String wikiPageName, TimestampType tt,
            double secsSinceEpoch) {
        String colName = mapTimestampTypeToColumnName(tt);

        connWrap.execSql("update wikiwordcontent set " + colName + " = ? " +
                "where word = ?", secsSinceEpoch, wikiPageName);
    }

}


/*


*/
