package de.mbutscher.wikiandpad.defaultwikilang;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.regex.Pattern;

import de.mbutscher.wikiandpad.WikiPage;
import de.mbutscher.wikiandpad.parsing.AbstractParserGrammar;
import de.mbutscher.wikiandpad.parsing.And;
import de.mbutscher.wikiandpad.parsing.Choice;
import de.mbutscher.wikiandpad.parsing.FindFirst;
import de.mbutscher.wikiandpad.parsing.Forward;
import de.mbutscher.wikiandpad.parsing.MatchFirst;
import de.mbutscher.wikiandpad.parsing.NonTerminalNode;
import de.mbutscher.wikiandpad.parsing.OptimizerOptions;
import de.mbutscher.wikiandpad.parsing.ParseAction;
import de.mbutscher.wikiandpad.parsing.ParseChoiceAction;
import de.mbutscher.wikiandpad.parsing.ParseException;
import de.mbutscher.wikiandpad.parsing.ParseStartAction;
import de.mbutscher.wikiandpad.parsing.ParserElement;
import de.mbutscher.wikiandpad.parsing.ParsingState;
import de.mbutscher.wikiandpad.parsing.Regex;
import de.mbutscher.wikiandpad.parsing.SyntaxNode;
import de.mbutscher.wikiandpad.parsing.TerminalNode;
import de.mbutscher.wikiandpad.utils.HashMapBuilder;
import de.mbutscher.wikiandpad.utils.StringOps;


/* It is essential to not use any capturing groups "()" in any regular
 * expression here. Only non-capturing groups "(?:)" are allowed!!!
 */

public class ParserGrammar extends AbstractParserGrammar {

    protected static final int RE_FLAGS = Pattern.DOTALL | Pattern.MULTILINE;
    
    protected static final ParseAction actionHideOnEmpty = new ParseAction() {
        public SyntaxNode parseAction(String s, int loc, ParsingState st,
                SyntaxNode t) throws ParseException {
            if (t.getFirstToken().getStrLength() == 0)
                return new NonTerminalNode();

            return null;
        }
    };

    protected static final String EscapePlainCharPAT = "\\\\";
    

    public static Regex buildRegex(String pattern, String name,
            boolean hideOnEmpty) {
        Regex element;

        if (name == null)
            element = regex(pattern, RE_FLAGS);
        else
            element = (Regex) regex(pattern, RE_FLAGS)
                    .setResultsNameNoCopy(name).setName(name);

        if (hideOnEmpty)
            element.setParseAction(actionHideOnEmpty);
            
        return element;
    }
    
    public static Regex buildRegex(String pattern, String name) {
        return buildRegex(pattern, name, false);
    }

    public static Regex buildRegex(String pattern) {
        return buildRegex(pattern, null, false);
    }
    
    /**
     * If using subpages this is used to resolve a link to the right wiki word
        relative to basePage on which the link is placed.
        It returns the absolute link (page name).
     * @param link
     * @param basePage
     * @return
     */
    protected static String resolveWikiWordLink(String link, WikiPage basePage) {
        return WikiLangHelper.resolvePrefixSilenceAndWikiWordLink(
                link, basePage).pageName;
    }

    protected static TerminalNode getFirstTerminalNode(NonTerminalNode t) {
        if (t.getChildrenCount() == 0)
            return null;
        
        SyntaxNode lt = t.get(-1);
        if (!(lt instanceof TerminalNode))
            return null;
            
        return (TerminalNode) lt;
    }

    protected static And text = null;

    protected static Forward content;
    protected static Forward headingContent;
    protected static Forward titleContent;
    protected static Forward characterAttributionContent;
    protected static Regex whitespace;
    protected static ParserElement stringEnd;
    
    protected static ParserElement escapedChar;
    protected static ParserElement bold;
    protected static ParserElement italics;
    protected static ParserElement boldEnd;
    protected static ParserElement italicsEnd;
    protected static Regex bracketStart;
    protected static Regex bracketEnd;
    protected static ParserElement title;
    
    protected static ParserElement wikiWord;
    protected static ParserElement urlRef;
    
    
    
    protected static ParseStartAction createCheckNotIn(String... tokNames) {
        final Set<String> tokNameSet = new HashSet<String>(Arrays.asList(tokNames));

        final ParseStartAction checkNoContain =
                new ParseStartAction() {
    
            public void parseStartAction(String s, int loc, ParsingState st,
                    ParserElement pe) throws ParseException {
                for (int i = 0; i < st.nameStack.size() - 1; ++i) {
                    if (tokNameSet.contains(st.nameStack.get(i))) {
                        throw new ParseException(s, loc,
                                String.format("Token '%s' is not allowed here",
                                st.nameStack.get(i)));
                    }
                }
            }
        };

        return checkNoContain;
    }

    
    protected static void buildLanguageSimpleFormatting() {
        
        escapedChar = buildRegex(EscapePlainCharPAT).and(
                buildRegex(".", "plainText"));

        ParserElement italicsStart = buildRegex("\\b_");
        italicsStart = italicsStart.setParseStartAction(createCheckNotIn(
                "italics"));

        italicsEnd = buildRegex("_\\b");

        italics = italicsStart.and(characterAttributionContent)
                .and(italicsEnd);
        italics = italics.setResultsNameNoCopy("italics").setName("italics");

        ParserElement boldStart = buildRegex("\\*(?=\\S)");
        boldStart = boldStart.setParseStartAction(createCheckNotIn("bold"));

        boldEnd = buildRegex("\\*");
        // boldEnd.inspect = "boldEnd";

        bold = boldStart.and(characterAttributionContent)
                .and(boldEnd);
        bold = bold.setResultsNameNoCopy("bold").setName("bold");
    }

    
    protected static Regex headingEnd;
    protected static ParserElement heading;

    protected static void buildLanguageHeading() {
        final ParseAction actionHeading = new ParseAction() {
            public SyntaxNode parseAction(String s, int loc,
                    ParsingState st, SyntaxNode t) throws ParseException {
                
                NonTerminalNode nt = (NonTerminalNode) t;
                nt.dPut("level", nt.getFirstToken().getStrLength());
                SyntaxNode contentNode = nt.findFlatByName("headingContent");
                nt.dPut("contentNode", contentNode);
                if (contentNode == null)
                    throw new ParseException(s, loc, "a heading needs content");

                return null;
            }
        };
        
        headingEnd = buildRegex("\\n");
        heading = buildRegex("^\\+{1,15}(?!\\+)").and(
                optional(buildRegex(" "))).and(headingContent).and(headingEnd);
        
        heading = heading.setResultsNameNoCopy("heading")
                .setParseAction(actionHeading);
    }

    protected static void buildLanguageWikiWord() {
        // WikiWords
        
        final ParseAction actionCutRightWhitespace = new ParseAction() {
            public SyntaxNode parseAction(String s, int loc,
                    ParsingState st, SyntaxNode t) throws ParseException {
                NonTerminalNode nt = (NonTerminalNode) t;
                TerminalNode lt = getFirstTerminalNode(nt);

                if (lt == null)
                    return null;
    
                String txt = lt.getText();

                for (int i = txt.length() - 1; i <= 0; --i) {
                    if ("\t \n\r".indexOf(txt.charAt(i)) == -1) {
                        if (i < txt.length() - 1) {
                            lt.setText(txt.substring(0, i + 1));
                            lt.recalcStrLength();
                            SyntaxNode t2 = SyntaxNode.buildSyntaxNode(
                                    txt.substring(i + 1), lt.getPos() + i + 1);
                            nt.add(t2);
                        }
                        return t;
                    }
                }
                return null;
            }
        };
        
        final ParseAction actionWikiWordNcc = new ParseAction() {
            public SyntaxNode parseAction(String s, int loc,
                    ParsingState st, SyntaxNode t) throws ParseException {

                NonTerminalNode nt = (NonTerminalNode) t;
                
                nt.dPut("wikiWord", nt.findFlatByName("word"));
                if (nt.dContainsKey("wikiWord")) {

                    String linkCore = ((SyntaxNode) nt.dGet("wikiWord"))
                            .getString();
                    
                    String pageName = WikiLangHelper
                            .resolvePrefixSilenceAndWikiWordLink(linkCore,
                                    (WikiPage) st.dictStack.get("baseWikiPage"))
                                    .pageName;

                    if (!StringOps.notEmpty(pageName)) {
                        throw new ParseException(s, loc,
                                "Subpage resolution of wikiword failed");
                    }
                    
                    // TODO: Test for Ncc blacklist
                    /*
        
                    if t.wikiWord in wikiFormatDetails.wikiDocument.getNccWordBlacklist():
                        raise ParseException(s, l, "Non-CamelCase word is in blacklist")
        
                    */
                    
                    nt.dPut("wikiWord", pageName);

                }
                nt.dPut("titleNode", nt.findFlatByName("title"));
                
                nt.dPut("fragmentNode", nt.findFlatByName("searchFragment"));
                if (nt.dGet("fragmentNode") != null) {
                    nt.dPut("searchFragment",
                            ((SyntaxNode) nt.dGet("fragmentNode"))
                            .dGet("unescaped"));
                }
                else
                    nt.dPut("searchFragment", null);
    
                nt.dPut("anchorLink", nt.findFlatByName("anchorLink"));
                if (nt.dGet("anchorLink") != null) {
                    nt.dPut("anchorLink",
                            ((SyntaxNode) nt.dGet("anchorLink")).getString());
                }
                
                return null;
            }

        };
    
        final ParseStartAction preActCheckWikiWordCcAllowed =
                new ParseStartAction() {
    
            public void parseStartAction(String s, int loc, ParsingState st,
                    ParserElement pe) throws ParseException {
                // TODO:
                /*
                try:
                    wikiFormatDetails = st.dictStack["wikiFormatDetails"]
                    
                    if not wikiFormatDetails.withCamelCase:
                        raise ParseException(s, l, "CamelCase words not allowed here")
                except KeyError:
                    pass
                */
            }
        };
    
        final ParseAction actionWikiWordCc = new ParseAction() {
            public SyntaxNode parseAction(String s, int loc,
                    ParsingState st, SyntaxNode t) throws ParseException {
                NonTerminalNode nt = (NonTerminalNode) t;
                
                nt.dPut("wikiWord", nt.findFlatByName("word"));
                if (nt.dContainsKey("wikiWord")) {

                    String linkCore = ((SyntaxNode) nt.dGet("wikiWord"))
                            .getString();
                    
                    String pageName = WikiLangHelper
                            .resolvePrefixSilenceAndWikiWordLink(linkCore,
                                    (WikiPage) st.dictStack.get("baseWikiPage"))
                                    .pageName;

                    if (!StringOps.notEmpty(pageName)) {
                        throw new ParseException(s, loc,
                                "Subpage resolution of wikiword failed");
                    }
                    
                    // TODO: Test for Cc blacklist
                    /*
        
                    if t.wikiWord in wikiFormatDetails.wikiDocument.getCcWordBlacklist():
                        raise ParseException(s, l, "CamelCase word is in blacklist")
        
                    */
                    
                    nt.dPut("wikiWord", pageName);
                
                }
                nt.dPut("titleNode", null);
                
                nt.dPut("fragmentNode", nt.findFlatByName("searchFragment"));
                if (nt.dGet("fragmentNode") != null) {
                    nt.dPut("searchFragment",
                            ((SyntaxNode) nt.dGet("fragmentNode"))
                            .dGet("unescaped"));
                }
                else
                    nt.dPut("searchFragment", null);
    
                nt.dPut("anchorLink", nt.findFlatByName("anchorLink"));
                if (nt.dGet("anchorLink") != null) {
                    nt.dPut("anchorLink",
                            ((SyntaxNode) nt.dGet("anchorLink")).getString());
                }
                
                return null;
            }
        };
    
        @SuppressWarnings("unused")
        String BracketStart = "[";
        String BracketStartPAT = "\\[";

        @SuppressWarnings("unused")
        String BracketEnd = "]";
        String BracketEndPAT = "\\]";
        
        String WikiWordPathPartPAT =
                "(?!\\.\\.)[^\\\\/\\[\\]\\|\\000-\\037=:;#!]+";
        String WikiPageNamePAT = WikiWordPathPartPAT + "(?:/" +
                WikiWordPathPartPAT + ")*";
    
        // Begins with dotted path parts which mean to go upward in subpage path
        String WikiWordDottedPathPAT = "\\.\\.(?:/\\.\\.)*(?:/" +
                WikiWordPathPartPAT + ")*";
        String WikiWordNonDottedPathPAT = "/{0,2}" + WikiPageNamePAT;
    
        String WikiWordNccPAT = WikiWordDottedPathPAT + "|" +
                WikiWordNonDottedPathPAT;
    
        String WikiWordTitleStartPAT = "\\|";
        @SuppressWarnings("unused")
        String WikiWordAnchorStart = "!";
        String WikiWordAnchorStartPAT = "!";
        
        String WikiWordCcPAT = "(?:[" +
                StringOps.UPPERCASE +
                "]+[" +
                StringOps.LOWERCASE +
                "]+[" +
                StringOps.UPPERCASE +
                "]+[" +
                StringOps.LETTERS + StringOps.DIGITS +
                "]*|[" +
                StringOps.UPPERCASE +
                "]{2,}[" +
                StringOps.LOWERCASE +
                "]+)";
        
        bracketStart = buildRegex(BracketStartPAT);
        bracketEnd = buildRegex(BracketEndPAT);
    
        Regex wikiWordNccCore = buildRegex(WikiWordNccPAT, "word");
    
        
        ParserElement searchFragmentExtern = buildRegex("#").and(
                buildRegex("(?:(?:#.)|[^ \\t\\n#])+", "searchFragment"));
                // .setParseAction(actionSearchFragmentExtern)
    
        ParserElement searchFragmentIntern = buildRegex("#").and(
                buildRegex("(?:(?:" + EscapePlainCharPAT +
                ".)|(?!" + WikiWordTitleStartPAT +
                "|" +  BracketEndPAT + ").)+", "searchFragment"));
                // .setParseAction(actionSearchFragmentIntern)
    
        ParserElement wikiWordAnchorLink = buildRegex(WikiWordAnchorStartPAT).and(
                buildRegex("[A-Za-z0-9_]+", "anchorLink"));
                
        title = buildRegex(WikiWordTitleStartPAT + "[ \\t]*")
                .and(titleContent);
    
                
        ParserElement wikiWordNcc = bracketStart.and(
                        wikiWordNccCore.copy().addParseAction(actionCutRightWhitespace)).and(
                        optional(matchFirst(searchFragmentIntern, wikiWordAnchorLink))).and(
                        whitespace).and(optional(title)).and(bracketEnd).and(
                        optional(matchFirst(searchFragmentExtern, wikiWordAnchorLink)));
    
        wikiWordNcc = wikiWordNcc.setResultsNameNoCopy("wikiWord")
                .setName("wikiWordNcc").setParseAction(actionWikiWordNcc);
    
        // wikiWordNcc.inspect = "wikiWordNcc";
        
        
        ParserElement wikiWordCc = buildRegex("\\b(?<!~)" + WikiWordCcPAT + 
                "\\b", "word").and(optional(matchFirst(searchFragmentExtern,
                wikiWordAnchorLink)));
        wikiWordCc = wikiWordCc.setResultsNameNoCopy("wikiWord")
                .setName("wikiWordCc")
                .setParseStartAction(preActCheckWikiWordCcAllowed)
                .setParseAction(actionWikiWordCc);
        
        
        // wikiWordCc.inspect = "wikiWordCc";
    
        wikiWord = wikiWordNcc.or(wikiWordCc);
        
        // wikiWord.inspect = "wikiWord";

    }
    
    protected static void buildLanguageUrl() {
        final ParseAction actionModeAppendix = new ParseAction() {
            public SyntaxNode parseAction(String s, int loc,
                    ParsingState st, SyntaxNode t) throws ParseException {

                NonTerminalNode nt = (NonTerminalNode) t;

                ArrayList<String[]> entries = new ArrayList<String[]>();
    
                for (SyntaxNode entry: nt.iterFlatByName("entry", 0)) {
                    NonTerminalNode nentry = (NonTerminalNode) entry;
                    String key = nentry.findFlatByName("key").getString();
                    String data = nentry.findFlatByName("data").getString();
                    entries.add(new String[] {key, data});
                }

                t.dPut("entries", entries);
                return t;
            }
        };
    
    
        final ParseAction actionUrlLink = new ParseAction() {
            public SyntaxNode parseAction(String s, int loc,
                    ParsingState st, SyntaxNode t) throws ParseException {
                
                NonTerminalNode nt = (NonTerminalNode) t;

                if ("urlLinkBare".equals(t.getName()))
                    t.dPut("bracketed", Boolean.FALSE);
                else
                    t.dPut("bracketed", Boolean.TRUE);
                
                t.setName("urlLink");        
                t.dPut("appendixNode", nt.findFlatByName("urlModeAppendix"));
                
                SyntaxNode coreNode = nt.findFlatByName("url");
                
                t.dPut("coreNode", coreNode);
    
                // Valid URL but may differ from original input
                t.dPut("url", StringOps.urlQuoteSpecific(coreNode.getString(),
                        " "));
                t.dPut("titleNode", nt.findFlatByName("title"));
                
                return null;
            }
        };

        ParserElement modeAppendixEntry = buildRegex("(?![;\\|\\]])\\S", "key")
                .and(buildRegex("(?:(?![;\\|\\]])\\S)*", "data"));
        
        modeAppendixEntry = modeAppendixEntry.setResultsNameNoCopy("entry");
        ParserElement modeAppendix = modeAppendixEntry.and(
                zeroOrMore(buildRegex(";").and(modeAppendixEntry)));
        
        modeAppendix = modeAppendix.addParseAction(actionModeAppendix);

        String UrlPAT =
                "(?:(?:https?|ftp|rel|wikirel)://|mailto:|Outlook:\\S|wiki:/|file:/)" +
                "(?:(?![.,;:!?)]+(?:[\"\\s]|$))[^\"\\s|\\]<>])*";


        String UrlInBracketsPAT =
                "(?:(?:https?|ftp|rel|wikirel)://|mailto:|Outlook:\\S|wiki:/|file:/)" + 
                "(?:(?![ \\t]+[|\\]])(?: |[^\"\\s|\\]<>]))*";

        
        ParserElement urlModeAppendix = modeAppendix.setResultsName(
                "urlModeAppendix");

        ParserElement urlWithAppend = buildRegex(UrlPAT, "url").and(
                optional(buildRegex(">").and(urlModeAppendix)));

        ParserElement urlWithAppendInBrackets = buildRegex(UrlInBracketsPAT,
                "url").and(optional(buildRegex(">").and(
                urlModeAppendix)));

        ParserElement urlBare = urlWithAppend.setResultsName("urlLinkBare");
        urlBare = urlBare.setParseAction(actionUrlLink);

        ParserElement urlTitled = bracketStart.and(urlWithAppendInBrackets)
                .and(whitespace).and(optional(title)).and(bracketEnd);

        urlTitled = urlTitled.setResultsNameNoCopy("urlLinkBracketed")
                .setParseAction(actionUrlLink);

        urlRef = urlTitled.or(urlBare);
    }
    
    protected static HashMap<String, ParserElement> TOKEN_TO_END;

    protected static void buildLanguage() {
        content = new Forward();
        headingContent = (Forward) (new Forward()).setResultsNameNoCopy(
                "headingContent");
        titleContent = (Forward) (new Forward()).setResultsNameNoCopy("title");
        characterAttributionContent = new Forward();

        whitespace = buildRegex("[ \t]*");
        whitespace = (Regex) whitespace.setParseAction(actionHideOnEmpty);

        stringEnd = buildRegex("(?!.)", "stringEnd");  // stringEnd();  // 


        final ParseAction pseudoActionFindMarkup = new ParseAction() {
            public SyntaxNode parseAction(String s, int loc,
                    ParsingState st, SyntaxNode ppd) throws ParseException {
                
                TerminalNode t = ppd.getFirstTokenT();
                
                if (t.getStrLength() == 0) {
                    return new NonTerminalNode();
                }
                t.setName("plainText");
                return t;
            }
        };
        
        buildLanguageSimpleFormatting();
        buildLanguageHeading();
        buildLanguageWikiWord();
        buildLanguageUrl();
        
        TOKEN_TO_END = (new HashMapBuilder<String, ParserElement>())
                .put("bold", boldEnd)
                .put("italics", italicsEnd)
                .put("wikiWord", bracketEnd)
                .put("urlLinkBracketed", bracketEnd)
                .put("heading", headingEnd)
                .getHashMap();
        
        final ParseChoiceAction chooseEndToken = new ParseChoiceAction() {
            public ParserElement parseChoiceAction(String s, int loc, ParsingState st,
                    ParserElement pe) throws ParseException {
                for(int i = st.nameStack.size() - 1; i >= 0 ; --i) {
                    ParserElement end = TOKEN_TO_END.get(st.nameStack.get(i));
                    if (end != null)
                        return end;
                }
                return stringEnd;
            }
        };
        
        ArrayList<ParserElement> endTokenCh = new ArrayList<ParserElement>();
        
        endTokenCh.add(stringEnd);
        endTokenCh.addAll(TOKEN_TO_END.values());
        
        
        Choice endToken = new Choice(endTokenCh.toArray(new ParserElement[0]),
                chooseEndToken);


        FindFirst findMarkupInHeading = findFirst(
                new ParserElement[] {bold, italics, escapedChar, urlRef,
                escapedChar,  wikiWord}, endToken);

        findMarkupInHeading = findMarkupInHeading
                .setPseudoParseAction(pseudoActionFindMarkup);
        
        headingContent.assign(zeroOrMore(
                notAny(endToken).and(findMarkupInHeading)));

        
        MatchFirst endTokenInCharacterAttribution = endToken.or(heading);

        FindFirst findMarkupInCharacterAttribution = findFirst(
                new ParserElement[] {bold, italics, escapedChar},
                endTokenInCharacterAttribution);

        findMarkupInCharacterAttribution = findMarkupInCharacterAttribution
                .setPseudoParseAction(pseudoActionFindMarkup);
        
        // findMarkup.inspect = "findMarkup";
        // bold.inspect = "bold";
        
        FindFirst findMarkup = findFirst(new ParserElement[] {bold, italics,
                urlRef, escapedChar, wikiWord, heading}, endToken);

        findMarkup = findMarkup.setPseudoParseAction(pseudoActionFindMarkup);

        content.assign(zeroOrMore(notAny(endToken).and(findMarkup)));
        
        characterAttributionContent.assign(zeroOrMore(
                notAny(endTokenInCharacterAttribution)
                .and(findMarkupInCharacterAttribution)));
         
        text = content.and(stringEnd);
        
        // text.inspect = "text";
        
        OptimizerOptions optOptions = new OptimizerOptions();
        optOptions.regexcombine = true;
        text.optimize(optOptions);
    }
    
    
    public static synchronized And getTextParserElement() {
        if (text == null)
            buildLanguage();

        return text;
    }


}



/*


    @staticmethod
    def parse(intLanguageName, content, formatDetails, threadstop):

        if len(content) == 0:
            return buildSyntaxNode([], 0, "text")

        if formatDetails.noFormat:
            return buildSyntaxNode([buildSyntaxNode(content, 0, "plainText")],
                    0, "text")

        baseDict = _buildBaseDict(formatDetails=formatDetails)

##         _prof.start()
        try:
            t = text.parseString(content, parseAll=True, baseDict=baseDict,
                    threadstop=threadstop)
            t = buildSyntaxNode(t, 0, "text")

            t = _TheParser._postProcessing(intLanguageName, content, formatDetails,
                    t, threadstop)

        finally:
##             _prof.stop()
            pass

        return t




*/






