package de.mbutscher.wikiandpad.parsing;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import de.mbutscher.wikiandpad.utils.StringOps;

public class RegexCombiner {
    // Match or search for the first matching expression
    public static final int REMODE_MATCH = 0;
    public static final int REMODE_SEARCH = 1;

    // Match or search for all matching expressions
    public static final int REMODE_MATCH_ALL = 2;
    public static final int REMODE_SEARCH_ALL = 3;
    
    protected ParserElement[] exprs;
    protected int reMode;
    protected int flags = -1;
    protected int flagsMask;
    protected String cleanPattern;
    protected Pattern regex;


    public RegexCombiner(ParserElement[] exprs, int reMode) {
        this.exprs = exprs;
        this.reMode = reMode;
    }
    
    public RegexCombiner(List<ParserElement> exprs, int reMode) {
        this(exprs.toArray(new ParserElement[exprs.size()]), reMode);
    }

    
    public ParserElement get(int i) {
        return exprs[i];
    }

    public int getFlags() {
        return flags;
    }

    public int getFlagsMask() {
        return flagsMask;
    }

    public String getCleanPattern() {
        return cleanPattern;
    }

    public Pattern getRegex() {
        return regex;
    }
    
    public static int[] combineRegexFlags(int flags, int flagsMask,
            int otherFlags, int otherFlagsMask) {
        if (flags == -1)
            return new int[] {otherFlags, otherFlagsMask};
        
        int commonMask = flagsMask & otherFlagsMask;
        int ownImportantFlags = flags & commonMask;
        int otherImportantFlags = otherFlags & commonMask;
        int newFlagsMask = flagsMask | otherFlagsMask;
        
        if (ownImportantFlags != otherImportantFlags)
            return new int[] {-1, newFlagsMask};
        
        return new int[] {(flags & flagsMask) | (otherFlags & otherFlagsMask),
                newFlagsMask};
    }
    
    protected boolean _compareCombineOtherFlags(int otherFlags,
            int otherFlagsMask) {
        
        int[] comb = combineRegexFlags(this.flags, this.flagsMask,
                otherFlags, otherFlagsMask);
        int flags = comb[0];
        int flagsMask = comb[1];
        
        if (flags == -1)
            return false;
        
        this.flags = flags;
        this.flagsMask = flagsMask;

        return true;
    }

    public boolean combine() {
        ArrayList<String> regexPatterns = new ArrayList<String>();
        
        if (exprs.length > 99)
            return false;
        
        for(int i = 0; i < exprs.length; ++i) {
            ParserElement exp = exprs[i];
            
            while (exp instanceof Forward) {
                exp = ((Forward) exp).expr;
            }
            
            if (exp instanceof FindFirst && reMode == REMODE_SEARCH) {
                RegexCombiner subCombiner = ((FindFirst) exp).getRegexCombiner();
                if (subCombiner == null)
                    return false;

                if (!_compareCombineOtherFlags(subCombiner.getFlags(),
                        subCombiner.getFlagsMask()))
                    return false;
                
                regexPatterns.add(subCombiner.getCleanPattern());
            }
            else if (exp instanceof NecessaryRegexProvider) {
                Pattern otherRe = ((NecessaryRegexProvider) exp).getRegex();
                if (otherRe == null)
                    return false;

                if (!_compareCombineOtherFlags(otherRe.flags(),
                        ((NecessaryRegexProvider) exp).getRegexFlagsMask()))
                    return false;
                
                regexPatterns.add(otherRe.pattern());
            }
            else
                return false;
        }
        
        cleanPattern = StringOps.strJoin("|", regexPatterns);
        
        if ((reMode & REMODE_MATCH_ALL) != 0) {
            for (int i = 0; i < regexPatterns.size(); ++i) {
                regexPatterns.set(i, "((?=" +
                        regexPatterns.get(i) + "))?");
            }
            String selectionPart = StringOps.strJoin(regexPatterns);
            
            if ((reMode & REMODE_SEARCH_ALL) != 0) {
                regex = Pattern.compile(selectionPart +
                        "(?:" + cleanPattern + ")", flags);
            }
            else
                regex = Pattern.compile(selectionPart,  flags);
            
            return true;
        }
        else {
            for (int i = 0; i < regexPatterns.size(); ++i) {
                regexPatterns.set(i, "(" +
                        regexPatterns.get(i) + ")");
            }
            regex = Pattern.compile(StringOps.strJoin("|", regexPatterns),
                    flags);
            
            return true;
        }
    }
    
    public static class FindAllResult {
        public int loc;
        public List<Integer> styles;

        public FindAllResult(int loc, List<Integer> styles) {
            this.loc = loc;
            this.styles = styles;
        }
    }
    

    public FindAllResult findAll(String instring, int start) {
        assert reMode == REMODE_SEARCH_ALL || reMode == REMODE_MATCH_ALL;
        
        instring = instring.substring(start);
        Matcher m = regex.matcher(instring);
        boolean success;

        if (reMode == REMODE_SEARCH_ALL) {
            success = m.find();
        }
        else {
            success = m.lookingAt();
        }
        
        if (!success)
            return new FindAllResult(start, Arrays.asList(new Integer[0]));
        
        ArrayList<Integer> styles = new ArrayList<Integer>();
        for(int i = 0; i < exprs.length; ++i) {
            String r = m.group(i + 1);
            if (r != null) {
                styles.add(i);
            }
        }
        
        if (styles.size() > 0)
            return new FindAllResult(m.start() + start, styles);
        
        if (reMode == REMODE_SEARCH_ALL)
            throw new RuntimeException(
                    "Internal error: No matching style in RegexCombiner (match all)");
        else
            return new FindAllResult(start, Arrays.asList(new Integer[0]));
    }
    
    
    @Override
    public String toString() {
        return String.format("<RegexCombiner %s>", regex.pattern());
    }


}
