package de.mbutscher.wikiandpad.utils;

import java.util.Collection;
import java.util.HashMap;

/** A ModelEvent informs about changes in the data of the application.
 * The usage differs from the normal java event handling:
 *
 * Each object which wants to send model events holds a public
 * ModelEvent object. To fire an event, it uses one of the
 * send() methods.
 *
 * TODO: Rewrite to use WeakReferences
 *
 * @author Michael Butscher
 */

public class MiscEvent implements Cloneable
{
    protected IdentityArrayList<MiscEventListener> listeners =
            new IdentityArrayList<MiscEventListener>();
    protected Object source = null;
    protected HashMap<String, Object> properties = null;
    protected MiscEvent parent = null;
    protected boolean compactionNeeded;
    protected int sendCallDeepness;

    /** Creates a new instance of ModelEvent */
    public MiscEvent() {
        properties = new HashMap<String, Object>();
        compactionNeeded = false;
        sendCallDeepness = 0;
    }

    /** Creates a new instance of ModelEvent */
    public MiscEvent(Object source) {
        this();
        this.source = source;
    }

    public Object getSource()
    {
        return source;
    }
    
    public void setSource(Object source) {
        this.source = source;
    }

    /** Return value for specified key or null if not found.
    Be careful: The value itself may be null. */
    @SuppressWarnings("unchecked")
    public <T> T getValue(String key)
    {
        return (T) properties.get(key);
    }

    /** Has the event the specified key? */
    public boolean hasKey(String key)
    {
        return properties.containsKey(key);
    }
    
    /** Has the event at least one of the keys in the collection? */
    public boolean hasAnyKey(Collection<String> possibleKeys)
    {
        for (String key: possibleKeys) {
            if (properties.containsKey(key))
                return true;
        }
        return false;
    }
    

    /** The ModelEvent which was called to fire this clone. If it returns null this is not a clone.*/
    public MiscEvent getParent()
    {
        return parent;
    }

    @Override
    @SuppressWarnings("unchecked")
    public MiscEvent clone()
    {
        MiscEvent result;
        try
        {
            result = (MiscEvent)super.clone();
        }
        catch (CloneNotSupportedException e)
        {
            e.printStackTrace();
            return null;
        }

        result.listeners = listeners;  // (ArrayList<MiscEventListener>) listeners.clone();
        if (properties != null)
            result.properties = (HashMap<String, Object>) properties.clone();

        return result;
    }
    
    public MiscEvent cloneIt() {
        MiscEvent result = clone();
        result.parent = this;
        return result;
    }
    
    
    public MiscEvent cloneIt(Object source) {
        MiscEvent result = cloneIt();
        result.setSource(source);
        return result;
    }


    /** A ModelEvent manages the listener list itself. */
    public synchronized void addListener(MiscEventListener listener)
    {
        if (!listeners.contains(listener)) {
            listeners.add(listener);
            if (listener instanceof ExtendedMiscEventListener)
                ((ExtendedMiscEventListener)listener).setHoldingEvent(this);
        }
    }

    public synchronized void removeListener(MiscEventListener listener)
    {
        int i = listeners.indexOf(listener);
        if (i != -1) {
            listeners.set(i, null);
            compactionNeeded = true;
        }
    }
    
    public static boolean isValid(MiscEventListener listener) {
        if (listener == null)
            return false;
        
        if (listener instanceof ExtendedMiscEventListener) {
            return ((ExtendedMiscEventListener)listener).isValid();
        }
        
        return true;
    }
    
    public synchronized int findListenerByKey(Object key, int start) {
        for (int i=start; i < listeners.size(); ++i) {
            MiscEventListener listener = listeners.get(i);
            if (isValid(listener) &&
                    (listener instanceof ExtendedMiscEventListener) &&
                    ((ExtendedMiscEventListener)listener).matchesListenerKey(key)) {
                return i;
            }
        }
        return -1;
    }
    
    public synchronized void removeListenersByKey(Object key)
    {
        int pos = 0;
        while (true) {
            pos = findListenerByKey(key, pos);
            if (pos == -1)
                break;

            listeners.set(pos, null);
            compactionNeeded = true;
            ++pos;
        }
    }

    
    public synchronized void setEnabledListenersByKey(Object key, boolean enabled)
    {
        int pos = 0;
        while (true) {
            pos = findListenerByKey(key, pos);
            if (pos == -1)
                break;
            
            try {
                ExtendedMiscEventListener listener =
                        (ExtendedMiscEventListener) listeners.get(pos);
                listener.setEnabled(enabled);
            }
            catch (ClassCastException e) {
                // TODO Programming error?
            }

            ++pos;
        }
    }

    /** While an event is sent listeners could be removed or became
     * invalid. compactify creates a new clean list without invalid
     * entries if necessary
     */
    public synchronized void compactify() {
        if (!compactionNeeded)
            return;
        
        IdentityArrayList<MiscEventListener> newListeners =
                new IdentityArrayList<MiscEventListener>();
        
        for(MiscEventListener listener: listeners) {
            if (isValid(listener))
                newListeners.add(listener);
        }
        listeners = newListeners;
        compactionNeeded = false;
    }


    /** Convenience method */ 
    public static void addListener(MiscEventSource source,
            MiscEventListener listener)
    {
        if (source == null)
            return;
        
        source.getMiscEvent().addListener(listener);
    }

    /** Convenience method */
    public static void removeListener(MiscEventSource source,
            MiscEventListener listener)
    {
        if (source == null)
            return;

        source.getMiscEvent().removeListener(listener);
    }
    

    /** Add a key-value pair to the internal Hashtable.
    <B>Can't be called on an original ModelEvent, must be a clone.</B>

    @return  this, so you can chain the call: event = event.put("a", a).put("foo", bar);

    @throws IllegalArgumentException   if this is not a clone  */

    public MiscEvent put(String key, Object value)
    {
        if (getParent() == null)
            throw new IllegalArgumentException("This is not a clone");  // TODO(MB) Create/Find a better exception

        properties.put(key, value);

        return this;
    }

    /** Remove a key from the internal property-table.
      <B>Can't be called on an original ModelEvent, must be a clone.</B>

      @return  this, so you can chain the call: event = event.put("a", a).put("foo", bar);

     */

    public MiscEvent remove(String key)
    {
        if (getParent() == null)
            throw new IllegalArgumentException("This is not a clone");  // TODO(MB) Create/Find a better exception
        properties.remove(key);

        return this;
    }

    /** Same as put(key, true). */
    public MiscEvent put(String key)
    {
        return put(key, Boolean.TRUE);
    }

    /** Called on the clone to dispatch itself to first, then to all listeners.
    <B>Can't be called on an original ModelEvent, must be a clone.</B>

    @param first   the first listener the event dispatches before dispatching to remaining listeners. A null value is ignored.
    @throws IllegalArgumentException   if this is not a clone  */

    public MiscEvent send(MiscEventListener first)
    {
        if (getParent() == null)
            throw new IllegalArgumentException("This is not a clone");   // TODO(MB) Create/Find a better exception
        if (first != null)
            first.miscEventHappened(this);

        // This function must be reentrant therefore call deepness is counted
        ++sendCallDeepness;
        try {
            MiscEventListener[] l = listeners.toArray(
                    new MiscEventListener[listeners.size()]);
            for (int i=0; i<l.length; ++i) {
                if (isValid(l[i]))
                    l[i].miscEventHappened(this);
            }
        }
        finally {
            --sendCallDeepness;
        }
        
        // If we are out of all send processing we can clean up
        if (sendCallDeepness == 0)
            compactify();
        
        return this;
    }

    /** Called on the clone to dispatch itself to all listeners.
    <B>Can't be called on an original ModelEvent, must be a clone.</B>

    @throws IllegalArgumentException  if this is not a clone  */

    public MiscEvent send()
    {
        return send(null);
    }

    /** Creates a clone with the appropriate data, so dispatching can be done later.<BR>
      Some methods can be called only on a cloned ModelEvent. To add properties, use the #put method.
      The parent of the cloned event is this event.

      @param _source   The object which will dispatch the event
      @param _cause    The cause (caller) of the event.  */

    public MiscEvent createClone(Object _source, Object _cause)
    {
        MiscEvent event = clone();
        if (event.properties == null)
            event.properties = new HashMap<String, Object>();

        event.source=_source;
        event.parent=this;

        return event;
    }

    /** Creates a clone with the appropriate data, so dispatching can be done later.<BR>
      Some methods can be called only on a cloned ModelEvent.
      The parent of the cloned event is this event.

    @param _source   The object which will dispatch the event
    @param _keys     Array of keys
    @param _values   Array of values for the keys. Its length must not be less than length of _keys array.
    @param _cause    The cause (caller) of the event.

    public MiscEvent createClone(MiscEventSource _source, String[] _keys, Object[] _values, Object _cause)
    {
        MiscEvent event=createClone(_source, _cause);

        for(int i = 0; i < _keys.length; ++i)
            event.put(_keys[i], _values[i]);

        return event;
    }

    public MiscEvent createClone(MiscEventSource _source, String[] _keys, Object _cause)
    {
        Object [] values = new Object [_keys.length];
        return createClone(_source, _keys, values, _cause);
    }

    public MiscEvent createClone(MiscEventSource _source, String _key, Object _value, Object _cause)
    {
        String [] keys = new String [] {_key};
        Object [] values = new Object [] {_value};
        return createClone(_source, keys, values, _cause);
    }

    public MiscEvent createClone(MiscEventSource _source, String _key, Object _cause)
    {
        String [] keys = new String [] {_key};
        Object [] values = new Object [1];
        return createClone(_source, keys, values, _cause);
    }
         */


    /** A ModelEvent sends a clone of itself to the listeners. */
    /*
    public void send(MiscEventSource _source, String[] _keys, Object[] _values, Object _cause)
    {
        createClone(_source, _keys, _values, _cause).send();
    }

    public void send(MiscEventSource _source, String[] _keys, Object _cause)
    {
        createClone(_source, _keys, _cause).send();
    }

    public void send(MiscEventSource _source, String _key, Object _value, Object _cause)
    {
        createClone(_source, _key, _value, _cause).send();
    }

    public void send(MiscEventSource _source, String _key, Object _cause)
    {
        createClone(_source, _key, _cause).send();
    }
    */
    



}
