package de.mbutscher.wikiandpad.utils.ui;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashSet;

import android.app.Activity;
import android.app.AlertDialog;
import android.content.res.Configuration;
import android.os.Bundle;
import android.os.Handler;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.HorizontalScrollView;
import android.widget.ImageButton;
import android.widget.LinearLayout;
import android.widget.ScrollView;
import android.widget.Toast;

public abstract class AbstractGuiHelper<AC extends Activity> {
    
    protected Handler uiHandler = null;

    public AbstractGuiHelper() {
    }
    
    public abstract <T extends AC> T getActivity();


    // ---------- Basic listener setting ----------
    
    public void setOnClickListener(int id, OnClickListener listener) {
        findViewById(id).setOnClickListener(listener);
    }
    
    
    protected static OnClickListener createOnClickListener(final Object obj,
            Method m) {
        final Method toCall = m;
        
        OnClickListener listener = null;
        if (toCall.getParameterTypes().length == 0) {
            listener = new OnClickListener() {

                public void onClick(View v) {
                    try {
                        toCall.invoke(obj);
                    }
                    catch(InvocationTargetException e) {
                        e.printStackTrace();
                    } catch (IllegalArgumentException e) {
                        e.printStackTrace();
                    } catch (IllegalAccessException e) {
                        e.printStackTrace();
                    }
                }
            };
        }
        else if (toCall.getParameterTypes().length == 1) {
            listener = new OnClickListener() {

                public void onClick(View v) {
                    try {
                        toCall.invoke(obj, v);
                    }
                    catch(InvocationTargetException e) {
                        e.printStackTrace();
                    } catch (IllegalArgumentException e) {
                        e.printStackTrace();
                    } catch (IllegalAccessException e) {
                        e.printStackTrace();
                    }
                }
            };
        }
        return listener;
    }

    /**
     * Create and set {@link OnClickListener} for view to call method
     * named listenerMethodName of obj
     * 
     * @param obj
     * @param view
     * @param listenerMethodName
     */
    public void setOnClickListener(Object obj, View view,
            String listenerMethodName) {
        for(Method m: obj.getClass().getMethods()) {
            if (listenerMethodName.equals(m.getName())) {
                OnClickListener listener = createOnClickListener(obj, m);
                
                view.setOnClickListener(listener);
                return;
            }
        }
        for(Method m: obj.getClass().getDeclaredMethods()) {
            if (listenerMethodName.equals(m.getName())) {
                m.setAccessible(true);

                OnClickListener listener = createOnClickListener(obj, m);
                
                view.setOnClickListener(listener);
                return;
            }            
        }
        
        throw new RuntimeException(String.format(
                "Method for ActivityGuiHelper.setOnClickListener " +
                "not found: %s", listenerMethodName));
    }

    /** For all methods in object with {@link ClickHandler} annotation
     * (which provides view id) the annotated method is called when
     * view is clicked by setting the OnClickListener
     * 
     * @param obj
     */
    public void setAllClickHandlers(Object obj) {
        HashSet<Integer> processedIds = new HashSet<Integer>();
        
        for(Method m: obj.getClass().getMethods()) {
            ClickHandler clickHandler = m.getAnnotation(ClickHandler.class);
            if (clickHandler == null)
                continue;
            
            int id = clickHandler.value();
            if (processedIds.contains(id))
                continue;
            
            processedIds.add(id);
           
            OnClickListener listener = createOnClickListener(obj, m);
            
            findViewById(id).setOnClickListener(listener);
        }
            
        for(Method m: obj.getClass().getDeclaredMethods()) {
            ClickHandler clickHandler = m.getAnnotation(ClickHandler.class);
            if (clickHandler == null)
                continue;
            
            int id = clickHandler.value();
            if (processedIds.contains(id))
                continue;
            
            processedIds.add(id);
           
            OnClickListener listener = createOnClickListener(obj, m);
            
            findViewById(id).setOnClickListener(listener);
        }
    }
    
    
    /**
     * Change a view and its subviews to landscape mode by some basic changes
     * if device is in landscape mode 
     * @param view View to change
     * @return view to replace the given one or same view if replacement
     * could be avoided
     */
    public View rewriteIfLandscape(View view) {
        View result = view;
        
        if (getActivity().getResources().getConfiguration().orientation !=
                Configuration.ORIENTATION_LANDSCAPE)
            return result;

        boolean switchLayoutParams = false;
        
        if (view instanceof Button) {
            switchLayoutParams = true;
        }
        else if (view instanceof ImageButton) {
            switchLayoutParams = true;
            int padLeft = view.getPaddingTop();
            int padRight = padLeft;
            int padTop = view.getPaddingBottom();
            int padBottom = padTop + view.getPaddingBottom() - view.getPaddingTop();
            view.setPadding(padLeft, padTop, padRight, padBottom);

        }
        else if (view instanceof HorizontalScrollView) {
            result = new ScrollView(view.getContext());
            result.setId(view.getId());
            switchLayoutParams = true;
        }
        else if (view instanceof LinearLayout) {
            LinearLayout ll = (LinearLayout) view;
            if (ll.getOrientation() == LinearLayout.HORIZONTAL)
                ll.setOrientation(LinearLayout.VERTICAL);
            else
                ll.setOrientation(LinearLayout.HORIZONTAL);
        }
        
        if (view instanceof ViewGroup) {
            ViewGroup vgIn = (ViewGroup) view;
            ViewGroup vgOut = (ViewGroup) result;

            View[] views = new View[vgIn.getChildCount()];
            
            for(int c = 0; c < vgIn.getChildCount(); ++c) {
                views[c] = vgIn.getChildAt(c);
            }
            
            vgIn.removeAllViews();
            vgOut.removeAllViews();

            for(View subView: views) {
                vgOut.addView(rewriteIfLandscape(subView));
            }
        }
        
        if (switchLayoutParams) {
            ViewGroup.LayoutParams params = view.getLayoutParams();
            int temp = params.height;
            params.height = params.width;
            params.width = temp;
            
            result.setLayoutParams(params);
        }
        
        if (result instanceof ImageButton) {
            ViewGroup.LayoutParams params = result.getLayoutParams();
            params.height = ViewGroup.LayoutParams.WRAP_CONTENT;
        }
        
        return result;
    }
    


    // ---------- Sub-bar handling ----------
    
    protected HashSet<Integer> subBarIds = null;

    public View.OnClickListener onBtnHideSubBars =
            new View.OnClickListener() {

        public void onClick(View v) {
            hideSubBars();
        }
    };

    public class OnClickShowSubBar implements View.OnClickListener {

        protected int id;


        public OnClickShowSubBar(int id) {
            this.id = id;
        }

        public void onClick(View v) {
            hideSubBars();
            findViewById(id).setVisibility(View.VISIBLE);
        }
        
    }

    public HashSet<Integer> getSubBarIds() {
        return subBarIds;
    }

    public void setSubBarIds(HashSet<Integer> subBarIds) {
        this.subBarIds = subBarIds;
    }

    
    public void hideSubBars() {
        if (subBarIds == null)
            return;

        for(int id: subBarIds) {
            View subBar = findViewById(id);
            subBar.setVisibility(View.GONE);
        }
    }
    
    
    public void saveSubBarVisibility(Bundle savedInstanceState) {
        for(int id: subBarIds) {
            savedInstanceState.putInt("viewId" + id + "_visibility",
                    findViewById(id).getVisibility());
        }
    }

    public void loadSubBarVisibility(Bundle savedInstanceState) {
        for(int id: subBarIds) {
            findViewById(id).setVisibility(savedInstanceState.getInt(
                    "viewId" + id + "_visibility", View.GONE));
        }
    }
    
    // ---------- Alert messages ----------
    
    public void alertInfo(String text) {
        Toast.makeText(getActivity(), text, Toast.LENGTH_SHORT).show();
    }
    
    public void alertInfo(String formatMessage, String... args) {
        alertInfo(String.format(formatMessage, (Object[]) args));
    }


    public void alertProblem(String text) {
        Toast.makeText(getActivity(), text, Toast.LENGTH_LONG).show();
    }

    public void alertProblem(String formatMessage, String... args) {
        alertProblem(String.format(formatMessage, (Object[]) args));
    }

    
    public void alertError(String text) {
        new AlertDialog.Builder(getActivity()).setMessage(text)
                .setPositiveButton("OK", null)
                .show();
    }
    
    public void alertError(String formatMessage, String... args) {
        alertError(String.format(formatMessage, (Object[]) args));
    }
    
    // ---------- Miscellanous ----------
    
    /* public View findViewById(int id) {
        return getActivity().findViewById(id);
    }
    
    
    public TextView findTextViewById(int id) {
        return (TextView) findViewById(id);
    } */
    
    @SuppressWarnings("unchecked")
    public <T extends View> T findViewById(int id) {
        return (T) getActivity().findViewById(id);
    }
    
    public Handler getUiHandler() {
        if (uiHandler == null)
            uiHandler = new Handler();
        
        return uiHandler;
    }
    
    public void runAfter(Runnable r) {
        getUiHandler().post(r);
    }

}
